@@ -29,6 +29,7 @@
 //! fn main() -> ees::MainResult {
 //!     do_work()?;
 //!     do_work().map_err(
+//!         // add additional context
 //!         |e| ees::wrap!(e, "failed to do work"))?;
 //!     Ok(())
 //! }
@@ -64,6 +65,7 @@ impl fmt::Display for ErrorChain<'_> {
 
 /// Print the complete error chain of an error, separated with colons
 #[must_use]
+#[inline]
 pub fn print_error_chain<'a>(error: impl error::Error + 'a) -> impl fmt::Display + 'a {
     ErrorChain {
         error: Box::new(error),
@@ -101,35 +103,39 @@ pub type MainResult = Result<(), MainError>;
 /// Construct an error on the fly
 #[macro_export]
 macro_rules! err {
-    ($($arg:tt)*) => {
-        $crate::internal::make_opaque($crate::internal::FormattedError {
-            message: std::format!($($arg)*),
-        })
-    }
+    ($fmt:expr) => {
+        $crate::internal::error_from_args(::std::format_args!($fmt))
+    };
+
+    ($fmt:expr, $($args:tt)*) => {
+        $crate::internal::error_from_args(::std::format_args!($fmt, $($args)*))
+    };
 }
 
 /// Construct an error on the fly, and immediately return from the current function
 #[macro_export]
 macro_rules! bail {
     ($($arg:tt)*) => {
-        return Err($crate::err!($($arg)*).into());
-    }
+        return Err(::std::convert::Into::into($crate::err!($($arg)*)));
+    };
 }
 
 /// Wrap an error in a new on-the-fly error
 #[macro_export]
 macro_rules! wrap {
-    ($source:expr, $($arg:tt)*) => {
-        $crate::internal::make_opaque($crate::internal::FormattedWrapError {
-            message: std::format!($($arg)*),
-            source: ($source).into(),
-        })
-    }
+    ($source:expr, $fmt:expr) => {
+        $crate::internal::wrap_error_from_args($source, ::std::format_args!($fmt))
+    };
+
+    ($source:expr, $fmt:expr, $($args:tt)*) => {
+        $crate::internal::wrap_error_from_args($source, ::std::format_args!($fmt, $($args)*))
+    };
 }
 
 /// Convert any error into a type that implements [std::error::Error]. This
 /// is mainly useful for converting [Error](crate::Error) types to `anyhow::Error`
 /// or similar.
+#[inline]
 pub fn to_err(error: impl Into<Error>) -> impl error::Error + Send + Sync + 'static {
     internal::WrapError {
         inner: error.into(),
@@ -149,12 +155,13 @@ mod tests {
     }
 
     #[test]
-    #[allow(deprecated)]
     fn messages() {
         let e = crate::err!("unknown error");
-        let e = crate::wrap!(e, "{}", "test");
+        let _e2 = crate::err!("unknown error {}{3}{1}{2}{1}", 7, 3, 5, 1);
+        let e = crate::wrap!(e, "te{}{}", "st", 1);
+        let e = crate::wrap!(e, "outer test");
         let printed = crate::print_error_chain(e);
-        assert_eq!(printed.to_string(), "test: unknown error");
+        assert_eq!(printed.to_string(), "outer test: test1: unknown error");
     }
 
     #[test]
@@ -174,7 +181,7 @@ mod tests {
 
     #[test]
     fn to_err_tests() {
-        let error = test_bail().unwrap_err();
+        let error: crate::Error = test_bail().unwrap_err();
         let actual_error = crate::to_err(error);
         let actual_error_2 = crate::to_err(actual_error);
         assert_eq!(
@@ -182,4 +189,26 @@ mod tests {
             "bailing"
         );
     }
+
+    #[test]
+    fn test_wrap_io_err() {
+        std::fs::File::open("hello")
+            .map_err(|e| wrap!(e, "error"))
+            .unwrap_err();
+    }
+
+    fn _return_wrap() -> Result<(), crate::Error> {
+        // test macro expansion by running:
+        // cargo +nightly rustc --profile=check -- -Zunstable-options --pretty=expanded --cfg test
+        let e = err!("hi");
+        Err(wrap!(e, "wrap"))?;
+        Ok(())
+    }
+
+    #[test]
+    fn unused_return_value() {
+        // both err!() and wrap!() should show a warning if they are unused
+        let e = crate::err!("unknown error {}", 7);
+        let _ = crate::wrap!(e, "unknown error {}", 7);
+    }
 }
